/***********************************************************************************************************************
 *
 * blueBill Resources - open source birding
 * Copyright (C) 2009-2011 by Tidalwave s.a.s. (http://www.tidalwave.it)
 *
 ***********************************************************************************************************************
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
 * an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the License for the
 * specific language governing permissions and limitations under the License.
 *
 ***********************************************************************************************************************
 *
 * WWW: http://bluebill.tidalwave.it
 * SCM: https://java.net/hg/bluebill~resources-src
 *
 **********************************************************************************************************************/
package it.tidalwave.bluebill.taxonomy.birds.col;

import it.tidalwave.bluebill.taxonomy.Taxon.Builder;
import it.tidalwave.bluebill.taxonomy.Taxon.Rank;
import it.tidalwave.util.Id;
import java.util.ArrayList;
import java.util.Collection;
import java.util.List;
import javax.annotation.Nonnull;
import org.openrdf.repository.Repository;
import it.tidalwave.bluebill.taxonomy.Taxonomy;
import it.tidalwave.bluebill.taxonomy.Taxon;
import it.tidalwave.bluebill.taxonomy.birds.BirdTaxonomyImporter;
import it.tidalwave.util.Initializer;
import java.io.FileNotFoundException;
import lombok.extern.slf4j.Slf4j;
import static it.tidalwave.bluebill.taxonomy.elmo.ElmoTaxonomyVocabulary.*;
import static it.tidalwave.bluebill.taxonomy.Taxon.Rank.*;

/***********************************************************************************************************************
 *
 * @author  Fabrizio Giudici
 * @version $Id$
 *
 **********************************************************************************************************************/
@Slf4j
public abstract class CatalogueOfLifeImporter extends BirdTaxonomyImporter
  {
    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    public static class IgnoreException extends Exception
      {
      }

    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    @Nonnull
    public Taxonomy run (final @Nonnull Repository repository,
                         final @Nonnull String taxonomyName)
      throws Exception
      {
        init(repository);
        initialize();

        final String idPrefix = ID_TAXONOMY_PREFIX + "CatalogueOfLife";
        final Taxonomy taxonomy = taxonomyManager.createTaxonomy().withId(new Id(idPrefix)).
                                                                   withDisplayNames(createDisplayable("en", "CatalogueOfLife")).
                                                                   inRepository(repository).
                                                                   build();
        final Taxon aves = createAves(taxonomy, idPrefix);

        final Rank[] ranks = new Rank[] { FAMILY, GENUS, SPECIES, SUBSPECIES };
        int processedCount = 0;
        final Collection<String> paths = getPaths();
        final int maxCount = Integer.parseInt(System.getProperty("maxTaxaProcessCount", "99999"));

        for (final String path : paths)
          {
            log.info("Processing {}/{}: {}...", new Object[] { ++processedCount, paths.size(), path });
            final List<Initializer<Taxon>> initializers = new ArrayList<Initializer<Taxon>>();

            final String id = taxonUniqueIdManager.findId(path);
            final String[] split = path.split("/");
            final int rankIndex = split.length - 1;
            final Rank rank = ranks[rankIndex];
            String scientificName = "";

            if (rankIndex <= 1)
              {
                scientificName = split[rankIndex];
              }
            else if (rankIndex == 2)
              {
                scientificName = split[1] + " " + split[2];
              }
            else
              {
                scientificName = split[1] + " " + split[2] + " " + split[3];
              }
//                         System.err.println(id + " // " + rank + " // " + split.length + " // " + path + " // ");

            try
              {
                Builder builder = findOrCreateSubTaxon(aves, "", id).withRank(rank).
                                                                     withScientificName(scientificName).
                                                                     withSpecificEpythet(split[rankIndex]).
                                                                     withInitializers(initializers);
                builder = process(path, scientificName, builder);
                final Taxon taxon = builder.build();
              }
            catch (IgnoreException e)
              {
                log.info("Ignoring: {}", path);
              }
            catch (FileNotFoundException e)
              {
                log.warn("Broken taxon: {}: {}", path, e.toString());
                brokenSpecies.add(e.getMessage());
              }
            catch (Exception e)
              {
                log.warn("Broken taxon: {}: {}", path, e.toString());
                log.warn("", e);
                brokenSpecies.add(e.getMessage());
              }
            
            if (processedCount >= maxCount)
              {
                log.warn("Stopping after {} items as requested.", maxCount);
                break;  
              }
          }

        taxonUniqueIdManager.close();

        return taxonomy;
      }

    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    @Nonnull
    protected Collection<String> getPaths()
      {
        return taxonUniqueIdManager.getPaths();
      }

    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    @Nonnull
    protected abstract Builder process (@Nonnull String path,
                                        @Nonnull String scientificName,
                                        @Nonnull Builder builder)
      throws IgnoreException, Exception;
  }
